<?php

namespace Drupal\commerce_vado\Form;

use Drupal\commerce_product\Entity\ProductVariationTypeInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Messenger\Messenger;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\field\Entity\FieldConfig;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Basic settings form for Commerce VADO.
 *
 * Provides configuration form to toggle fields on product variation types.
 *
 * Class VadoManagementForm
 *
 * @package Drupal\commerce_vado\Form
 */
class VadoManagementForm extends ConfigFormBase {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * The messenger.
   *
   * @var \Drupal\Core\Messenger\Messenger
   */
  protected $messenger;

  /**
   * VadoManagementForm constructor.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The config factory.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entity_field_manager
   *   The entity field manager.
   * @param \Drupal\Core\Messenger\Messenger $messenger
   *   The messenger.
   */
  public function __construct(ConfigFactoryInterface $config_factory, EntityTypeManagerInterface $entity_type_manager, EntityFieldManagerInterface $entity_field_manager, Messenger $messenger) {
    parent::__construct($config_factory);
    $this->entityTypeManager = $entity_type_manager;
    $this->entityFieldManager = $entity_field_manager;
    $this->messenger = $messenger;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('config.factory'),
      $container->get('entity_type.manager'),
      $container->get('entity_field.manager'),
      $container->get('messenger')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'vado_management_form';
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames() {
    return ['commerce_vado.settings'];
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $wrapper = 'product_variations';
    $form['#prefix'] = '<div id="' . $wrapper . '">';
    $form['#suffix'] = '</div>';
    $config = $this->configFactory()->getEditable('commerce_vado.settings');

    $form['#tree'] = TRUE;

    $form['variation_types'] = [
      '#type' => 'details',
      '#title' => $this->t('Variation Types'),
      '#open' => TRUE,
      'warning' => ['#markup' => $this->t('<strong>WARNING:</strong> Unchecking a box and saving the form will cause all data for the field to be deleted!')]
    ];

    $variation_type_storage = $this->entityTypeManager->getStorage('commerce_product_variation_type');
    // Load & format product variation types for checkboxes.
    $variation_types = $variation_type_storage->loadMultiple();
    /** @var \Drupal\commerce_product\Entity\ProductVariationTypeInterface $variation_type */
    foreach ($variation_types as $variation_type) {
      $form['variation_types'][$variation_type->id()] = [
        '#type' => 'details',
        '#title' => $variation_type->label(),
        '#open' => $this->isVadoEnabled($variation_type),
        'fields' => $this->buildVadoFieldCheckboxElements($variation_type),
      ];
    }

    $form['settings_container'] = [
      '#type' => 'details',
      '#title' => $this->t('Settings'),
      '#open' => TRUE,
    ];

    $form['settings_container']['hide_parent_zero_price'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Hide Zero Price'),
      '#description' => $this->t('Hide parent variation price in cart and order views when set to zero.'),
      '#default_value' => $config->get('hide_parent_zero_price'),
      '#ajax' => [
        'callback' => [$this, 'ajaxWarning'],
        'wrapper' => $wrapper,
      ],
    ];

    $form['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Save'),
      '#button_type' => 'primary',
    ];

    return $form;
  }

  /**
   * Ajax callback for hide_parent_zero_price element.
   */
  public function ajaxWarning(array &$form, FormStateInterface $form_state) {
    $triggering_element = $form_state->getTriggeringElement();
    if ($triggering_element && end($triggering_element['#parents']) == 'hide_parent_zero_price') {
      $this->messenger->addWarning($this->t('You must clear the caches for this change to take effect.'));
    }

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    parent::validateForm($form, $form_state);
    $values = $form_state->getValues();

    // A variation type is considered enabled if one of the primary fields is enabled.
    $enabled_variation_types = [];
    foreach ($values['variation_types'] as $variation_type_id => $settings) {
      foreach ($this->getPrimaryFields() as $primary_field) {
        $is_enabled = (bool) $settings['fields'][$primary_field];
        if ($is_enabled) {
          $enabled_variation_types[] = $variation_type_id;
          continue 2;
        }
      }
    }

    // If a variation type is not enabled, ensure all fields are disabled.
    foreach ($values['variation_types'] as $variation_type_id => $settings) {
      if (!in_array($variation_type_id, $enabled_variation_types)) {
        foreach ($settings['fields'] as $field_name => $value) {
          $form_state->setValueForElement($form['variation_types'][$variation_type_id]['fields'][$field_name], 0);
        }
      }
    }

  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $values = $form_state->getValues();
    $form_display_storage = $this->entityTypeManager->getStorage('entity_form_display');

    $vado_fields = $this->getVadoFields();
    foreach ($values['variation_types'] as $variation_type_id => $settings) {
      $form_display = $form_display_storage->load('commerce_product_variation.' . $variation_type_id . '.default');
      foreach ($settings['fields'] as $field_name => $enabled) {
        $field_config = FieldConfig::loadByName('commerce_product_variation', $variation_type_id, $field_name);
        // Delete fields that are not enabled and are currently installed on the variation type.
        if (!$enabled && $field_config) {
          $field_config->delete();
        }
        // Create fields that are enabled and are currently NOT installed on the variation type.
        elseif ($enabled && !$field_config) {
          $field = FieldConfig::create([
            'entity_type' => 'commerce_product_variation',
            'field_name' => $field_name,
            'bundle' => $variation_type_id,
            'label' => $vado_fields[$field_name]['label'],
            'description' => $vado_fields[$field_name]['description'],
            'settings' => $vado_fields[$field_name]['settings'],
          ]);
          $field->save();

          // If the default form display does not exist yet for the variation type, create it.
          if (!$form_display) {
            $form_display = $form_display_storage->create([
              'status' => TRUE,
              'id' => 'commerce_product_variation.' . $variation_type_id . '.default',
              'targetEntityType' => 'commerce_product_variation',
              'bundle' => $variation_type_id,
              'mode' => 'default',
              'settings' => $vado_fields[$field_name]['settings'],
            ]);
          }
          // Set the form display settings for the variation type so the new fields are visible.
          $form_display->setComponent($field_name, [
            'label' => 'hidden',
            'type' => $vado_fields[$field_name]['form_widget'],
            'weight' => $vado_fields[$field_name]['weight'],
            'settings' => $vado_fields[$field_name]['settings'],
          ])->save();
        }
      }
    }

    $config = $this->configFactory()->getEditable('commerce_vado.settings');
    $config->set('hide_parent_zero_price', $values['settings_container']['hide_parent_zero_price']);
    $config->save();

    $this->messenger()->addMessage($this->t('The configuration has been updated.'));
  }

  /**
   * Gets the vado fields.
   *
   * @return array[]
   *   The vado fields.
   */
  protected function getVadoFields() {
    return [
      'child_variations' => [
        'label' => $this->t('Variation Add-on Children'),
        'description' => $this->t('Choose child variation(s) to be added to the cart with this parent.'),
        'field_type' => 'entity_reference',
        'settings' => [
          'target_type' => 'commerce_product_variation',
        ],
        'cardinality' => -1,
        'form_widget' => 'entity_reference_autocomplete',
        'weight' => 90,
        'primary_field' => TRUE,
      ],
      'variation_groups' => [
        'label' => $this->t('Variation Add-on Groups'),
        'description' => $this->t('Choose variation group(s) to be added to the cart with this parent.'),
        'field_type' => 'entity_reference',
        'settings' => [
          'target_type' => 'commerce_vado_group',
        ],
        'cardinality' => -1,
        'form_widget' => 'entity_reference_autocomplete',
        'weight' => 90,
        'primary_field' => TRUE,
      ],
      'sync_quantity' => [
        'label' => $this->t('Sync Cart Quantity'),
        'description' => $this->t('Lock the cart quanity of the selected child variation(s) to the parent quantity.'),
        'field_type' => 'boolean',
        'settings' => [],
        'cardinality' => 1,
        'form_widget' => 'boolean_checkbox',
        'weight' => 91,
        'primary_field' => FALSE,
        'dependencies' => [
          'child_variations',
          'variation_groups',
        ],
      ],
      'bundle_discount' => [
        'label' => $this->t('Bundle Discount'),
        'description' => $this->t('Enter a percentage to discount the price of the child variation(s) or variation group(s) for this parent.'),
        'field_type' => 'decimal',
        'settings' => [
          'max' => 100,
          'suffix' => '%',
        ],
        'cardinality' => 1,
        'form_widget' => 'number',
        'weight' => 92,
        'primary_field' => FALSE,
        'dependencies' => [
          'child_variations',
          'variation_groups',
        ],
      ],
        'include_parent' => [
          'label' => $this->t('Include Parent'),
          'description' => $this->t('Include the parent variation in the bundle discount applied to the child variation(s) or variation group(s).'),
          'field_type' => 'boolean',
        'settings' => [],
        'cardinality' => 1,
        'form_widget' => 'boolean_checkbox',
        'weight' => 93,
        'primary_field' => FALSE,
        'dependencies' => [
          'child_variations',
          'variation_groups',
        ],
      ],
    ];
  }

  /**
   * Gets the primary fields.
   *
   * Primary fields are the fields used to determine if
   * vado is enabled for a given product variation type.
   *
   * @return array
   *   The primary fields.
   */
  protected function getPrimaryFields() {
    $primary_fields = [];
    foreach ($this->getVadoFields() as $field_name => $settings) {
      if ($settings['primary_field']) {
        $primary_fields[] = $field_name;
      }
    }

    return $primary_fields;
  }

  /**
   * Determines if vado is enabled for the given variation type.
   *
   * Vado is enabled for a variation type if one of the primary
   * vado fields is enabled.
   *
   * @param \Drupal\commerce_product\Entity\ProductVariationTypeInterface $variation_type
   *   The variation type.
   * @return bool
   *   TRUE if vado is enabled for the variation type, FALSE otherwise.
   */
  protected function isVadoEnabled(ProductVariationTypeInterface $variation_type) {
    foreach ($this->getPrimaryFields() as $primary_field) {
      if ($this->hasField($primary_field, $variation_type)) {
        return TRUE;
      }
    }

    return FALSE;
  }

  /**
   * Builds the checkbox elements for each vado field
   * for the provided variation type.
   *
   * @param \Drupal\commerce_product\Entity\ProductVariationTypeInterface $variation_type
   *   The variation type.
   * @return array
   *   The checkbox elements.
   */
  protected function buildVadoFieldCheckboxElements(ProductVariationTypeInterface $variation_type) {
    $build = [];

    $vado_fields = $this->getVadoFields();
    foreach ($vado_fields as $field_name => $settings) {
      $build[$field_name] = [
        '#type' => 'checkbox',
        '#title' => $settings['label'],
        '#description' => $settings['description'],
        '#default_value' => $this->hasField($field_name, $variation_type),
        '#disabled' => $this->shouldDisableFieldCheckbox($field_name, $variation_type),
      ];

      // Fields with dependencies should not be visible unless a dependency is met.
      if (!empty($settings['dependencies'])) {
        foreach($settings['dependencies'] as $dependent_field_name) {
          $selector = ':input[name="variation_types['.$variation_type->id().'][fields]['.$dependent_field_name.']"]';
          $build[$field_name]['#states']['visible'][] = [$selector => ['checked' => TRUE]];
        }
      }
    }

    return $build;
  }

  /**
   * Determines whether the provided fields checkbox
   * should be disabled for the provided variation type.
   *
   * Primary fields should be protected so that if they have data
   * they cannot be uninstalled.
   *
   * @param $field_name
   *   The field name.
   * @param \Drupal\commerce_product\Entity\ProductVariationTypeInterface $variation_type
   *   The variation type.
   * @return bool
   *   TRUE if checkbox should be disabled, FALSE otherwise.
   */
  protected function shouldDisableFieldCheckbox($field_name, ProductVariationTypeInterface $variation_type) {
    return $this->hasFieldData($field_name, $variation_type) && in_array($field_name, $this->getPrimaryFields());
  }

  /**
   * Checks whether a variation type has a field.
   *
   * @param $field_name
   *   The field name.
   * @param \Drupal\commerce_product\Entity\ProductVariationTypeInterface $variation_type
   *   The variation type.
   * @return bool
   *   TRUE if the variation type has the provided field, FALSE otherwise.
   */
  protected function hasField($field_name, ProductVariationTypeInterface $variation_type) {
    $has_field = FALSE;

    $definitions = $this->entityFieldManager->getFieldDefinitions('commerce_product_variation', $variation_type->id());
    if (isset($definitions[$field_name])) {
      $has_field = TRUE;
    }

    return $has_field;
  }

  /**
   * Checks whether a variation type has data for a field.
   *
   * @param $field_name
   *   The field name.
   * @param \Drupal\commerce_product\Entity\ProductVariationTypeInterface $variation_type
   *   The variation type.
   * @return bool
   *   TRUE if the variation type has data for the provided field, FALSE otherwise.
   */
  protected function hasFieldData($field_name, ProductVariationTypeInterface $variation_type) {
    $has_field_data = FALSE;

    if (!$this->hasField($field_name, $variation_type)) {
      return $has_field_data;
    }

    $variation_storage = $this->entityTypeManager->getStorage('commerce_product_variation');
    $count = $variation_storage->getQuery()
      ->condition('type', $variation_type->id())
      ->exists($field_name)
      ->range(0, 1)
      ->count()
      ->execute();

    if ($count > 0) {
      $has_field_data = TRUE;
    }

    return $has_field_data;
  }

}
