<?php

namespace Drupal\commerce_vado\Plugin\Commerce\VadoGroupWidget;

use Drupal\commerce\EntityHelper;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\PluginBase;

/**
 * Provides the base class for vado group widgets.
 */
abstract class VadoGroupWidgetBase extends PluginBase implements VadoGroupWidgetInterface {

  /**
   * The group the plugin instance belongs to.
   *
   * @var \Drupal\commerce_vado\Entity\VadoGroupInterface
   */
  protected $parentEntity;

  /**
   * Constructs a VadoGroupWidgetBase object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);

    $this->setConfiguration($configuration);
  }

  /**
   * {@inheritdoc}
   */
  public function setParentEntity(EntityInterface $parent_entity) {
    $this->parentEntity = $parent_entity;
  }

  /**
   * {@inheritdoc}
   */
  public function getLabel() {
    return $this->pluginDefinition['label'];
  }

  /**
   * {@inheritdoc}
   */
  public function allowsMultiple() {
    return $this->pluginDefinition['allows_multiple'];
  }

  /**
   * {@inheritdoc}
   */
  public function buildElement(array $form, FormStateInterface $form_state) {
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function getDefaultValue() {
    $has_default_items = $this->parentEntity->hasDefaultItems();
    $default_items = $this->parentEntity->getDefaultItems();
    // multi-value elements expect an array of ids or an empty array.
    if ($this->allowsMultiple()) {
      return $has_default_items ? EntityHelper::extractIds($default_items) : [];
    }
    // single value elements expect a single id or NULL;
    else {
      return $has_default_items ? reset($default_items)->id() : NULL;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function calculateDependencies() {
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'group_title_display' => $this->getDefaultGroupTitleDisplayOption(),
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getConfiguration() {
    return $this->configuration;
  }

  /**
   * {@inheritdoc}
   */
  public function setConfiguration(array $configuration) {
    $this->configuration = NestedArray::mergeDeep($this->defaultConfiguration(), $configuration);
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form['group_title_display'] = [
      '#type' => 'select',
      '#title' => $this->t('Group Title Display'),
      '#options' => $this->buildGroupTitleDisplayOptions(),
      '#default_value' => $this->configuration['group_title_display'],
      '#required' => TRUE,
    ];
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state) { }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    if (!$form_state->getErrors()) {
      $values = $form_state->getValue($form['#parents']);
      $this->setConfiguration($values);
    }
  }

  /**
   * Gets the default group title display option.
   *
   * @return string
   *  The default group title display option.
   */
  protected function getDefaultGroupTitleDisplayOption() {
    return 'before';
  }

  /**
   * Builds the group title display options.
   *
   * @return array
   *   The group title display options array.
   */
  protected function buildGroupTitleDisplayOptions() {
    return [
      'before' => $this->t('Visible'),
      'invisible' => $this->t('Invisible')
    ];
  }

  /**
   * @return bool
   */
  protected function showEmptyOption() {
    return !$this->parentEntity->hasDefaultItems() || !$this->parentEntity->isRequired();
  }

  /**
   * @return bool
   */
  protected function shouldBuildEmptyOption() {
    return $this->showEmptyOption();
  }

  /**
   * @return array
   */
  protected function buildEmptyOption() {
    return ['' => $this->t('- None -')];
  }

  /**
   * @return array
   */
  protected function buildGroupItemOptions() {
    $addon_group_options = [];

    if ($this->shouldBuildEmptyOption()) {
      $addon_group_options += $this->buildEmptyOption();
    }
    foreach($this->parentEntity->getItems() as $group_item) {
      $addon_group_options[$group_item->id()] = $group_item->label();
    }

    return $addon_group_options;
  }

}
