<?php

namespace Drupal\commerce_vado\Resolver;

use Drupal\commerce\Context;
use Drupal\commerce\EntityHelper;
use Drupal\commerce\PurchasableEntityInterface;
use Drupal\commerce_price\Resolver\PriceResolverInterface;
use Drupal\commerce_price\RounderInterface;
use Drupal\commerce_vado\Entity\VadoGroupItem;
use Drupal\commerce_vado\VadoHelper;

class VadoPriceResolver implements PriceResolverInterface {

  /**
   * The price rounder.
   *
   * @var \Drupal\commerce_price\RounderInterface
   */
  protected $rounder;

  /**
   * Constructs a VadoPriceResolver.
   *
   * @param \Drupal\commerce_price\RounderInterface $rounder
   *   The price rounder.
   */
  public function __construct(RounderInterface $rounder) {
    $this->rounder = $rounder;
  }

  /**
   * {@inheritdoc}
   */
  public function resolve(PurchasableEntityInterface $entity, $quantity, Context $context) {
    /** @var \Drupal\commerce_product\Entity\ProductVariationInterface $parent_variation */
    $parent_variation = $entity;

    $price = NULL;
    // Price resolvers do not have access to the cart data, therefore selected group items
    // must be determined by getting the information directly from the request parameters.
    $request = \Drupal::request();
    $selected_group_item_ids = $request->request->get('commerce_vado_group', []);

    // Whether AJAX was just triggered by a vado group selection.
    $triggering_element_name = $request->request->get('_triggering_element_name', '');
    $vado_ajax_detected = str_contains($triggering_element_name, 'commerce_vado_group[addon_group_');

    // When the page first loads, no group items will be selected in the request parameters, so grab the
    // default values that are populated in the group widget. If a vado group selection just triggered AJAX
    // and no selections are present, do not load the defaults because the user has opted to not select any addons.
    if (!$vado_ajax_detected && empty($selected_group_item_ids) && $parent_variation->hasField('variation_groups')) {
      $groups = $parent_variation->get('variation_groups')->referencedEntities();
      /** @var \Drupal\commerce_vado\Entity\VadoGroupInterface $group */
      foreach ($groups as $group) {
        // Get the default value from the widget is more reliable in cases such as static list
        // where all items are added by default regardless of a group item being default.
        $selected_group_item_ids[] = $group->getGroupWidget()->getDefaultValue();
      }
    }

    // flattens the array of group item ids.
    if(!empty($selected_group_item_ids)) {
      $selected_group_item_ids = VadoHelper::processGroupSelections($selected_group_item_ids);
    }

    // create a multiplier value from the bundle discount integer value.
    $bundle_discount_integer = 0;
    if ($parent_variation->hasField('bundle_discount') && !$parent_variation->get('bundle_discount')->isEmpty()) {
      $bundle_discount_integer = $parent_variation->get('bundle_discount')->value;
    }
    $bundle_discount_multiplier = (100 - $bundle_discount_integer) / 100;

    // whether or not to include the parent variation in the bundle discount.
    $include_parent = $parent_variation->hasField('include_parent');
    if ($include_parent && !$parent_variation->get('include_parent')->isEmpty()) {
      $include_parent = $parent_variation->get('include_parent')->value;
    }

    // calculate the price from the variation addons, if present.
    if ($parent_variation->hasField('child_variations') && !$parent_variation->get('child_variations')->isEmpty()) {
      $price = $parent_variation->getPrice()->multiply($include_parent ? $bundle_discount_multiplier : 1);
      /** @var \Drupal\commerce\PurchasableEntityInterface $addon_variation */
      foreach ($parent_variation->get('child_variations')->referencedEntities() as $addon_variation) {
        $addon_price = $addon_variation->getPrice()->multiply($bundle_discount_multiplier);
        $addon_price = $this->rounder->round($addon_price);
        $price = $price->add($addon_price);
      }
    }

    // Calculate price from selected group items.
    if (!empty($selected_group_item_ids)) {
      // use the price calculated from the variation addons if it exists.
      if (!$price) {
        $price = $parent_variation->getPrice()->multiply($include_parent ? $bundle_discount_multiplier : 1);
      }

      $group_items = VadoGroupItem::loadMultiple($selected_group_item_ids);
      if (!empty($group_items)) {
        /** @var \Drupal\commerce_vado\Entity\VadoGroupItemInterface $group_item */
        foreach ($group_items as $group_item) {
          // Use the group_discount on the group item if available. If not, use the bundle_discount.
          $group = $group_item->getGroup();
          $group_discount_multiplier = $bundle_discount_multiplier;
          if ($group->hasGroupDiscount()) {
            $group_discount_multiplier = $group->getGroupDiscountMultiplier();
          }
          // Use the group_item_discount on the group item if available. If not, use the group_discount.
          $group_item_discount_multiplier = $group_discount_multiplier;
          if ($group_item->hasGroupItemDiscount()) {
            $group_item_discount_multiplier = $group_item->getGroupItemDiscountMultiplier();
          }
          // Get the group item variation and apply the discount.
          $variation = $group_item->getVariation();
          $variation_price = $variation->getPrice()->multiply($group_item_discount_multiplier);
          $variation_price = $this->rounder->round($variation_price);
          $price = $price->add($variation_price);
        }
      }
    }

    return $price;
  }
}
